//===================================RISCKER====================================
//============================Interruption Controller===========================
//------------------------------------------------------------------------------
//File Name:	interrupt_controller.v
//Purpose:		Manage all interruptions and deliver the vector for the ISR
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------

module interrupt_controller(	
		input clk,
		//FSM CONTROL
		input intoggle,  //Interruption State Toggles when Interruption is recognized and when user returns out of it
		input ipc_write, //PC is saved when Interruption is recognized
		//PORT IN
		input int_ext1, //External Interrupt
		input int_ext2, //External Interrupt
		//TIMERS
		input int_timer1,
		input int_timer2,
		//REG_FILE
		input [7:0]rc, //Register Control Flags
				//Interruption Routines Addresses
		input [15:0]ti1,  //Timer1 Interruption Routine Address
		input [15:0]ti2,  //Timer2 Interruption Routine Address
		input [15:0]eir1, //External Interruption Routine Address 1
		input [15:0]eir2, //External Interruption Routine Address 2
		
		//OUTPUTS
		output reg interruption=0, //Interruption Indicator
		output reg [15:0]int_addr,	//Interruption Routine Address
		output reg retfi=0 //Return from interruption
    );

//////////////////////////////////////////////////////////////////////////
//Control Register Flags 
//////////////////////////////////////////////////////////////////////////
wire reset_flag_int_timer1	=rc[0]; 
wire reset_flag_int_timer2	=rc[1];
wire reset_flag_int_ext1	=rc[2];
wire reset_flag_int_ext2	=rc[3];
wire enable_int_tr1			=rc[4];
wire enable_int_tr2			=rc[5];
wire enable_int_ext1		=rc[6];
wire enable_int_ext2		=rc[7];

//////////////////////////////////////////////////////////////////////////
//Interruption Flags Registers
//////////////////////////////////////////////////////////////////////////
reg flag_int_t2=0, //Timer2 Interruption
	 flag_int_t1=0, //Timer1 Interruption
	 flag_int_ext1=0, //External Interruption 1
	 flag_int_ext2=0; //External Interruption 2
	 
//////////////////////////////////////////////////////////////////////////	  
//Timer Interruption 
//////////////////////////////////////////////////////////////////////////

always@(posedge int_timer2, posedge reset_flag_int_timer2)
begin
		if (reset_flag_int_timer2)
			flag_int_t2<=1'b0;
		else
				if (enable_int_tr2)
				flag_int_t2<=1'b1;
				else
				flag_int_t2<=flag_int_t2;
end


always@(posedge int_timer1, posedge reset_flag_int_timer1)
begin
		if (reset_flag_int_timer1)
			flag_int_t1<=1'b0;
		else
				if (enable_int_tr1)
				flag_int_t1<=1'b1;
				else
				flag_int_t1<=flag_int_t1;
end

//////////////////////////////////////////////////////////////////////////
//External Interruptions
//////////////////////////////////////////////////////////////////////////
reg 	reg_int_ext1=0, reg_int_ext2=0; //External Interruptions Synchronization

always@(posedge clk)
reg_int_ext1 <= int_ext1;

always@(posedge clk)
reg_int_ext2 <= int_ext2;

//External Interruption 1
always@(posedge reg_int_ext1, posedge reset_flag_int_ext1)
begin
		if (reset_flag_int_ext1)
			flag_int_ext1<=1'b0;
		else	
			if (enable_int_ext1)
				flag_int_ext1<=1'b1;
			else
				flag_int_ext1<=flag_int_ext1;		
end

//External Interruption 1
always@(posedge reg_int_ext2, posedge reset_flag_int_ext2)
begin
		if (reset_flag_int_ext2)
			flag_int_ext2<=1'b0;
		else	
				if (enable_int_ext2)
					flag_int_ext2<=1'b1;
				else
					flag_int_ext2<=flag_int_ext2;
end
//////////////////////////////////////////////////////////////////////////
//Interruption State	 
//////////////////////////////////////////////////////////////////////////

//Global Interruption Enable "Toggle" Register
reg global_en_int = 0;

always@(posedge intoggle)
global_en_int<=~global_en_int;

//Interruptions State
reg [3:0] int_state=0;  

always @ (posedge clk)
if (global_en_int)
	int_state <= 0;
else
	int_state <= {flag_int_ext2,flag_int_ext1,flag_int_t2,flag_int_t1};

wire or_interrupt = |int_state;

always@(posedge or_interrupt, posedge ipc_write)
	if(ipc_write)	
		interruption<=1'b0;
	else
		interruption<=1'b1;

//Return from Interruption Register
wire retfi_clk = reset_flag_int_timer1|reset_flag_int_timer2|reset_flag_int_ext1|reset_flag_int_ext2;

always@(posedge retfi_clk, posedge intoggle)
	if (intoggle)
		retfi <= 0;
	else
		retfi <= 1;

//////////////////////////////////////////////////////////////////////////
//Interruption Service Routine Address Selector
//////////////////////////////////////////////////////////////////////////
	always@*
		if(flag_int_ext2)
			int_addr=eir2;
		else if(flag_int_ext1)
			int_addr=eir1;
		else if (flag_int_t2)
			int_addr=ti2;
		else if (flag_int_t1)
			int_addr=ti1;
		else 
			int_addr=16'd32;
////////////////////////////////////////////////////////////////////////// 
endmodule
